package redis

import (
	"context"
	"encoding/json"
	"flag"
	"fmt"
	"regexp"
	"strconv"
	"strings"
	"time"

	"github.com/go-redis/redis/v7"
	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"
)

var log = _log.New("redis")

var confTimeout = 3 * time.Second
var confInterval = time.Duration(0)

func InitFlags(fs *flag.FlagSet) {
	fs.DurationVar(&confTimeout, "redis.timeout", confTimeout, "[redis] operation timeout")
	fs.DurationVar(&confInterval, "redis.interval", confInterval, "[redis] interval between polls")
}

// FIXME: why is this a script?
var scriptGetAccounts = redis.NewScript(`
    return redis.call("hgetall", "accounts:")
`)

// FIXME: why is this a script?
var scriptGetContCount = redis.NewScript(`
    local key = "containers:" .. KEYS[1]
    return redis.call("ZCOUNT", key, 0, 0)
`)

var scriptAcctInfo = redis.NewScript(`
	local accts = redis.call("hgetall", "accounts:")
	local res = {}
	local index = 0
	for i, acc in ipairs(accts) do
        if i % 2 == 1 then
                local acct_key = "account:" .. acc;
                res[index] = {acc, redis.call('HGET', acct_key, 'bytes'), redis.call("HGET", acct_key, 'objects')}
                index = index + 1
        end;
	end;
	return cjson.encode(res);
`)

var Factory = &redisDaemonFactory{}

type redisDaemonFactory struct{}

func (df *redisDaemonFactory) Daemons(inventory *common.Inventory) map[string]interface{} {
	params := make(map[string]interface{})
	namespace := inventory.Namespace

	redisServices := inventory.Services["redis"]
	for _, service := range redisServices {
		params[service.ID] = &redisDaemonParams{
			Namespace:   namespace,
			ServiceID:   service.ID,
			ServiceType: "redis",
			ServiceIP:   service.IP,
			ServicePort: service.Port,
		}
	}

	redissentinelServices := inventory.Services["redissentinel"]
	for _, service := range redissentinelServices {
		params[service.ID] = &redisDaemonParams{
			Namespace:   namespace,
			ServiceID:   service.ID,
			ServiceType: "redissentinel",
			ServiceIP:   service.IP,
			ServicePort: service.Port,
		}
	}

	return params
}

func (df *redisDaemonFactory) New(paramsI interface{}) daemon.DaemonI {
	params := paramsI.(*redisDaemonParams)

	return &redisDaemon{
		namespace:   params.Namespace,
		serviceID:   params.ServiceID,
		serviceType: params.ServiceType,
		serviceIP:   params.ServiceIP,
		servicePort: params.ServicePort,
	}
}

type redisDaemonParams struct {
	Namespace   string
	ServiceID   string
	ServiceType string
	ServiceIP   string
	ServicePort int
}

type redisDaemon struct {
	daemon.BaseDaemon

	namespace   string
	serviceID   string
	serviceType string
	serviceIP   string
	servicePort int
	serviceAddr string

	client *redis.Client
}

func (d *redisDaemon) Interval() time.Duration {
	return confInterval
}

var dbRegex = regexp.MustCompile(`^db(\d+):keys=(\d+),expires=(\d+)`)

func (d *redisDaemon) Init(ctx context.Context) error {
	serviceAddr := fmt.Sprintf("%s:%d", d.serviceIP, d.servicePort)

	d.serviceAddr = serviceAddr
	d.client = redis.NewClient(&redis.Options{
		Addr:         serviceAddr,
		DialTimeout:  confTimeout,
		ReadTimeout:  confTimeout,
		WriteTimeout: confTimeout,
	}).WithContext(ctx)

	return nil
}

func (d *redisDaemon) Collect(ctx context.Context, m *common.MetricData) error {
	errStats := d.collectStats(m)

	if d.serviceType == "redissentinel" {
		return errStats
	}

	val, err := d.client.HLen("accounts:").Result()
	if err != nil {
		return err
	}
	if val <= 0 {
		return errStats
	}

	errAccount := d.collectAccount(m)

	if errStats != nil {
		return errStats
	}
	return errAccount
}

func (d *redisDaemon) sendStat(metric string, value float64, labels [][]string, m *common.MetricData) {
	m.Set(&common.Metric{
		Name:   "redis_" + metric,
		Labels: labels,
		Value:  value,
		TTL:    confInterval * 2,
	})
}

func (d *redisDaemon) collectStats(m *common.MetricData) error {
	data, err := d.client.Info().Result()
	if err != nil {
		m.IncrError("redis", "stats")
		return err
	}
	for _, line := range strings.Split(data, "\r\n") {
		kv := strings.SplitN(line, ":", 2)
		if len(kv) != 2 {
			continue
		}

		stat := kv[0]
		valStr := kv[1]

		labels := [][]string{
			[]string{"namespace", d.namespace},
			[]string{"service", d.serviceID},
			[]string{"service_type", d.serviceType},
			[]string{"service_addr", d.serviceAddr},
		}

		if matches := dbRegex.FindStringSubmatch(line); len(matches) == 4 {
			valKeys, errKeys := strconv.ParseFloat(matches[2], 64)
			valExpires, errExpires := strconv.ParseFloat(matches[3], 64)
			if errKeys != nil || errExpires != nil {
				// TODO log
				continue
			}
			labels = append(labels, []string{"db", matches[1]})

			labelsKeys := make([][]string, len(labels), len(labels)+1)
			copy(labelsKeys, labels)
			d.sendStat("db_keys", valKeys, append(labelsKeys, []string{"type", "keys"}), m)

			labelsExpires := make([][]string, len(labels), len(labels)+1)
			copy(labelsExpires, labels)
			d.sendStat("db_keys", valExpires, append(labelsExpires, []string{"type", "expires"}), m)
			continue
		}

		val, errFloat := strconv.ParseFloat(valStr, 64)

		switch stat {
		case "redis_mode":
			stat = "is_" + valStr
			val = 1
			errFloat = nil

		case "uptime_in_seconds":
			stat = "uptime_seconds"

		case "connected_clients":
			stat = "clients"
			labels = append(labels, []string{"type", "connected"})
		case "blocked_clients":
			stat = "clients"
			labels = append(labels, []string{"type", "connected"})

		case "used_memory":
			stat = "memory_bytes"
			labels = append(labels, []string{"type", "total"})
		case "used_memory_rss":
			stat = "memory_bytes"
			labels = append(labels, []string{"type", "rss"})
		case "used_memory_overhead":
			stat = "memory_bytes"
			labels = append(labels, []string{"type", "overhead"})
		case "used_memory_dataset":
			stat = "memory_bytes"
			labels = append(labels, []string{"type", "dataset"})
		case "used_memory_lua":
			stat = "memory_bytes"
			labels = append(labels, []string{"type", "lua"})
		case "used_memory_scripts":
			stat = "memory_bytes"
			labels = append(labels, []string{"type", "scripts"})
		case "used_memory_peak":
			stat = "used_memory_bytes_max"

		case "mem_fragmentation_ratio":
			stat = "memory_fragmentation_ratio"

		case "active_defrag_running":
			stat = "active_defrag_is_running"
		case "active_defrag_hits":
			stat = "active_defrag"
			labels = append(labels, []string{"type", "hits"})
		case "active_defrag_misses":
			stat = "active_defrag"
			labels = append(labels, []string{"type", "misses"})
		case "active_defrag_key_hits":
			stat = "active_defrag"
			labels = append(labels, []string{"type", "key_hits"})
		case "active_defrag_key_misses":
			stat = "active_defrag"
			labels = append(labels, []string{"type", "key_misses"})

		case "lazyfree_pending_objects":

		case "rdb_changes_since_last_save":
		case "rdb_bgsave_in_progress":
		case "rdb_last_save_time":
			stat += "_seconds"
		case "rdb_last_bgsave_status":
			if valStr == "ok" {
				val = 1.0
			} else {
				val = 0.0
			}
			errFloat = nil
		case "rdb_last_bgsave_time_sec":
			stat = "rdb_last_bgsave_time_seconds"
		case "rdb_current_bgsave_time_sec":
			stat = "rdb_current_bgsave_time_seconds"
		case "rdb_last_cow_size":
			stat += "_bytes"

		case "aof_enabled":
		case "aof_rewrite_in_progress":
		case "aof_rewrite_scheduled":
		case "aof_last_rewrite_time_sec":
			stat = "aof_last_rewrite_time_seconds"
		case "aof_current_rewrite_time_sec":
			stat = "aof_current_rewrite_time_seconds"
		case "aof_last_bgrewrite_status":
			if valStr == "ok" {
				val = 1.0
			} else {
				val = 0.0
			}
			errFloat = nil
		case "aof_last_write_status":
			if valStr == "ok" {
				val = 1.0
			} else {
				val = 0.0
			}
			errFloat = nil
		case "aof_last_cow_size":
			stat += "_bytes"
		case "aof_current_size":
			stat += "_bytes"
		case "aof_base_size":
			stat += "_bytes"
		case "aof_pending_rewrite":
		case "aof_buffer_length":
		case "aof_rewrite_buffer_length":
		case "aof_pending_bio_fsync":
		case "aof_delayed_fsync":
			stat += "_total"

		case "loading":
		case "loading_start_time":
			stat += "_seconds"
		case "loading_total_bytes":
		case "loading_loaded_bytes":
		case "loading_eta_seconds":

		case "rejected_connections":
			stat = "rejected_connections_total"
		case "total_connections_received":
			stat = "connections_total"
		case "total_commands_processed":
			stat = "commands_total"
		case "instantaneous_ops_per_sec":
			stat = "commands_ops"

		case "total_net_input_bytes":
			stat = "io_bytes_total"
			labels = append(labels, []string{"direction", "in"})
		case "total_net_output_bytes":
			stat = "io_bytes_total"
			labels = append(labels, []string{"direction", "out"})
		case "instantaneous_input_kbps":
			stat = "io_kbps"
			labels = append(labels, []string{"direction", "in"})
		case "instantaneous_output_kbps":
			stat = "io_kbps"
			labels = append(labels, []string{"direction", "out"})

		case "sync_full":
		case "sync_partial_ok":
			stat = "sync_partial"
			labels = append(labels, []string{"status", "ok"})
		case "sync_partial_err":
			stat = "sync_partial"
			labels = append(labels, []string{"status", "err"})

		case "expired_keys":
			stat = "keys_total"
			labels = append(labels, []string{"type", "expired"})
		case "evicted_keys":
			stat = "keys_total"
			labels = append(labels, []string{"type", "evicted"})

		case "keyspace_hits":
			stat = "keyspace_total"
			labels = append(labels, []string{"type", "hits"})
		case "keyspace_misses":
			stat = "keyspace_total"
			labels = append(labels, []string{"type", "misses"})

		case "pubsub_channels":
		case "pubsub_patterns":
		case "latest_fork_usec":
			stat = "latest_fork_duration_seconds"
			val /= 1e6
		case "migrate_cached_sockets":
		case "slave_expires_tracked_keys":
		case "role":
			labels = append(labels, []string{"role", valStr})
			val = 1
			errFloat = nil
		case "repl_backlog_active":
		case "repl_backlog_size":
			stat += "_bytes"
		case "master_link_status":
			labels = append(labels, []string{"status", valStr})
			val = 1
			errFloat = nil
		case "master_last_io_seconds_ago":
			stat = "master_last_io_seconds"
		case "master_sync_in_progress":
		case "slave_priority":
		case "slave_read_only":
		case "master_sync_left_bytes":
		case "master_sync_last_io_seconds_ago":
			stat = "master_sync_last_io_seconds"
		case "master_link_down_since_seconds":
		case "connected_slaves":
		case "min_slaves_good_slaves":

		default:
			/* ignore unknown commands */
			continue
		}

		if errFloat != nil {
			// TODO log
			continue
		}
		d.sendStat(stat, val, labels, m)
	}

	return nil
}

func (d *redisDaemon) collectAccount(m *common.MetricData) error {
	acctInfo, err := scriptAcctInfo.Run(d.client, []string{}, 0).Result()
	if err != nil {
		m.IncrError("redis", "account")
		return err
	}
	acctObj := map[string][]string{}
	err = json.Unmarshal([]byte(acctInfo.(string)), &acctObj)
	if err != nil {
		m.IncrError("redis", "account")
		return err
	}
	for _, data := range acctObj {
		labels := [][]string{
			[]string{"namespace", d.namespace},
			[]string{"service", d.serviceID},
			[]string{"service_type", d.serviceType},
			[]string{"service_addr", d.serviceAddr},
		}
		valSize, err := strconv.ParseFloat(data[1], 64)
		if err != nil {
			m.IncrError("redis", "account")
			log.Warnln("error while collecting account objects", d.serviceID, err)
			continue
		}
		valObj, err := strconv.ParseFloat(data[2], 64)
		if err != nil {
			m.IncrError("redis", "account")
			log.Warnln("error while collecting account objects", d.serviceID, err)
			continue
		}
		labels = append(labels, []string{"account", data[0]})
		m.Set(&common.Metric{
			Name:   "redis_account_size_bytes",
			Labels: labels,
			Value:  valSize,
			TTL:    confInterval * 2,
		})
		m.Set(&common.Metric{
			Name:   "redis_account_objects",
			Labels: labels,
			Value:  valObj,
			TTL:    confInterval * 2,
		})
	}

	accounts, err := scriptGetAccounts.Run(d.client, []string{}, 0).Result()
	if err != nil {
		m.IncrError("redis", "account")
		return err
	}
	for _, acctI := range accounts.([]interface{}) {
		acct, ok := acctI.(string)
		if !ok || acct == "1" {
			continue
		}
		count, err := scriptGetContCount.Run(d.client, []string{acct}, 1).Result()
		if err != nil {
			m.IncrError("redis", "account")
			log.Warnln("error while collecting account containers", acct, err)
			continue
		}
		m.Set(&common.Metric{
			Name: "redis_account_containers",
			Labels: [][]string{
				[]string{"namespace", d.namespace},
				[]string{"service", d.serviceID},
				[]string{"service_type", d.serviceType},
				[]string{"service_addr", d.serviceAddr},
				[]string{"account", acct},
			},
			Value: float64(count.(int64)),
			TTL:   confInterval * 2,
		})
	}

	return nil
}

func (d *redisDaemon) Close() error {
	return d.client.Close()
}
